from collections.abc import Mapping
import logging.config
from pathlib import Path

from box import Box
from dotenv import load_dotenv
from envyaml import EnvYAML

_logger = logging.getLogger(__name__)


def recursive_update(d, u):
    """Updates a dictionary d with updates from u recursively."""
    for k, v in u.items():
        if isinstance(v, Mapping):
            d[k] = recursive_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def load_config(config_path, include_environment=True, include_default_config=True):
    _logger.debug(f"Loading configuration file: {config_path}")
    print(f"Loading configuration file: {config_path}")
    if Path('.env').is_file():
        _logger.warning('Existence of a .env file may produce unexpected side-effects when loading the configuration'
                        'with EnvYAML. Use for example a .env.general file instead.')
        print('WARNING: Existence of a .env file may produce unexpected side-effects when loading the configuration'
              '         with EnvYAML. Use for example a .env.general file instead.')

    config = Box(EnvYAML(config_path, include_environment=include_environment).export())

    if include_default_config:
        default_logging_config = Box(EnvYAML(Path(__file__).parent / 'default_config.yml',
                                             include_environment=False).export())
        config = recursive_update(default_logging_config, config)

    return config


def start_script(dotenv='.env.general',
                 config_yaml='config.yaml',
                 include_environment=True,
                 include_default_config=True):
    load_dotenv(dotenv)
    config = load_config(config_yaml, include_environment=include_environment, include_default_config=include_default_config)
    Path(config.logging.handlers.debug_file_handler.filename).parent.mkdir(parents=True, exist_ok=True)
    Path(config.logging.handlers.info_file_handler.filename).parent.mkdir(parents=True, exist_ok=True)
    Path(config.logging.handlers.warning_file_handler.filename).parent.mkdir(parents=True, exist_ok=True)
    Path(config.logging.handlers.error_file_handler.filename).parent.mkdir(parents=True, exist_ok=True)
    Path(config.logging.handlers.critical_file_handler.filename).parent.mkdir(parents=True, exist_ok=True)
    logging.config.dictConfig(config.logging)
    return config



import logging
from pathlib import Path

import mlflow

from .tracking import get_model

_logger = logging.getLogger(__name__)


def get_registered_model(model_name: str,
                         model_stage: str):
    _logger.debug("Finding trained model on MLflow Server")
    _logger.debug(f"Model name: {model_name}")
    _logger.debug(f"Model stage: {model_stage}")

    client = mlflow.tracking.MlflowClient()
    registered_models = client.search_model_versions(f"name='{model_name}'")

    registered_models_in_stage = [rm for rm in registered_models
                                  if rm.current_stage == model_stage]

    if len(registered_models_in_stage) == 0:
        raise LookupError(f"{model_stage} version of model {model_name} not found")
    if len(registered_models_in_stage) > 1:
        _logger.warning(f"{len(registered_models_in_stage)} {model_stage} versions of model {model_name} found")

    rm = registered_models_in_stage[-1]
    _logger.info(f"Loading model '{model_name}' version {rm.version}")

    model_path = None
    for p in Path(rm.source).parents:
        if p.stem == 'artifacts':
            model_path = Path(rm.source).relative_to(p)

    if model_path is None:
        raise ValueError("Path of model in artifact store not found")

    return get_model(rm.run_id, str(model_path)), rm.version


def register_model(run_id: str,
                   model_name: str,
                   artifact_path: str = 'model'):
    _logger.debug("Registering model in MLflow Server")
    model_version = mlflow.register_model(
        f"runs:/{run_id}/artifacts/{artifact_path}",
        model_name
    )
    return model_version



import logging

import numpy as np
import pandas as pd
from sklearn.metrics import roc_auc_score, average_precision_score, roc_curve, precision_recall_curve

from myautoml.visualisation.colors import TEST_COLOR, BASELINE_COLOR

_logger = logging.getLogger(__name__)


def plot_roc(ax, y, y_pred_proba, legend_loc='best', label=None, *args, **kwargs):
    auc = roc_auc_score(y, y_pred_proba)
    fpr, tpr, threshold = roc_curve(y, y_pred_proba)

    ax.set_title('Receiver Operating Characteristic Curve')
    ax.set_xlabel('False Positive Rate')
    ax.set_ylabel('True Positive Rate')
    ax.set_xlim([-0.05, 1.05])
    ax.set_ylim([-0.05, 1.05])

    # Baseline / Random predictions
    ax.plot([0, 1], [0, 1], linestyle='dotted', color=BASELINE_COLOR)

    if label is None:
        plot_label = "area = {:0.2f}".format(auc)
    else:
        plot_label = "{}: area = {:0.2f}".format(label, auc)

    ax.plot(fpr, tpr, label=plot_label, *args, **kwargs)
    ax.legend(loc=legend_loc)
    return ax


def plot_precision_recall(ax, y, y_pred_proba, legend_loc='best', label=None, *args, **kwargs):
    auc = average_precision_score(y, y_pred_proba)
    precision, recall, threshold = precision_recall_curve(y, y_pred_proba)

    ax.set_title('Precision-Recall Curve')
    ax.set_xlabel('Recall')
    ax.set_ylabel('Precision')
    ax.set_xlim([-0.05, 1.05])
    ax.set_ylim([-0.05, 1.05])

    if label is None:
        plot_label = "area = {:0.2f}".format(auc)
    else:
        plot_label = "{}: area = {:0.2f}".format(label, auc)

    ax.plot(recall, precision, label=plot_label, *args, **kwargs)
    ax.legend(loc=legend_loc)
    return ax


def plot_lift_deciles(ax, y, y_pred_proba):
    # Sorting the class 1 probabilities and labels
    df = pd.DataFrame({'labels': y, 'pred': y_pred_proba})
    df = df.sort_values(by=['pred'], ascending=False)

    # Calculating baseline for class 1 occurrence
    baseline = df['labels'].sum() / len(df['labels'])

    # Binning into deciles and calculating actual class 1 percentages
    df_bins = df.groupby(pd.qcut(df['pred'], 10, duplicates='drop')).apply(
        lambda a: a['labels'].sum() / len(a['labels']))
    df_bins = (df_bins / baseline)

    # Plotting the chart
    ax.set_title('Non-Cumulative Lift')

    x = np.arange(len(df_bins)) + 1
    y = np.flip(df_bins.values)
    bars = ax.bar(x, y, color=TEST_COLOR)

    ax.axhline(y=1, color=BASELINE_COLOR, linestyle='dotted')

    for bar in bars:
        height = bar.get_height()
        ax.text(bar.get_x() + bar.get_width() / 2., height + 0.02, round(height, 2), ha='center', va='bottom')

    return ax


# noinspection SpellCheckingInspection
def _get_y_lims_for_range(r):
    d = (r[1] - r[0]) * 0.05
    return [r[0] - d, r[1] + d]


def plot_cum_precision(ax, y, y_pred_proba, label=None, color=TEST_COLOR):
    # Sorting the class 1 probabilities and labels
    df = pd.DataFrame({'labels': y, 'pred': y_pred_proba})
    df = df.sort_values(by=['pred'], ascending=False)

    # Calculating baseline for class 1 occurrence
    baseline = df['labels'].sum() / len(df['labels'])

    # Computing lift curve
    df['cumsum'] = df['labels'].cumsum()
    df['one'] = 1
    df['rank'] = df['one'].cumsum()
    df['fraction'] = df['rank'] / len(df['labels'])
    df['lift'] = df['cumsum'] / df['rank']

    # Plotting the graph
    ax.set_title('Cumulative Precision')
    ax.plot(df['fraction'], df['lift'], label=label, color=color)
    ax.axhline(y=baseline, color=BASELINE_COLOR, linestyle='dotted')
    ax.legend(loc="upper right")

    ax.set_xlim([-0.05, 1.05])
    ax.set_ylim([-0.05, 1.05])

    return ax


def plot_sorted_probabilities(ax, y, y_pred_proba, color=TEST_COLOR, ascending=True, label=None):
    # Sorting the class 1 probabilities and labels
    df = pd.DataFrame({'labels': y, 'pred_proba': y_pred_proba})
    df = df.sort_values(by=['pred_proba'], ascending=ascending)

    df['one'] = 1
    df['rank'] = df['one'].cumsum()
    df['fraction'] = df['rank'] / len(df['labels'])

    # Plotting the graph
    ax.set_title('Sorted probabilities')
    ax.plot(df['fraction'], df['pred_proba'], color=color, label=label)
    ax.legend(loc="upper center")
    return ax


def plot_prediction_distribution(ax, y_pred_proba, *args, **kwargs):
    ax.hist(y_pred_proba, bins=20, density=True, *args, **kwargs)
    ax.legend(loc='upper right')
    return ax



import numpy as np


def plot_grouped_bar(ax, groups_of_bars, x_labels, group_labels, colors):
    """Plot a grouped bar chart"""
    n_bars_in_group = len(groups_of_bars[0])
    bar_width = 1 / (n_bars_in_group + 1)

    x_pos = []
    for group_num, group in enumerate(groups_of_bars):
        x_pos = [x + group_num * bar_width for x in np.arange(len(group))]
        bars = ax.bar(x_pos, group, color=colors[group_num], width=bar_width, label=group_labels[group_num])

        for bar in bars:
            height = bar.get_height()
            ax.text(bar.get_x() + bar.get_width() / 2., height + 0.02, round(height, 2), ha='center', va='bottom')

    ax.set_xticks(x_pos, x_labels)
    ax.legend()

    return ax



from matplotlib import cm

TRAIN_COLORS = cm.get_cmap('Pastel1')
TEST_COLORS = cm.get_cmap('Set1')

EVALUATION_COLORS = {
    'train': 'lightsteelblue',
    'valid': 'steelblue',
    'test': 'navy',
    'baseline': 'lightgray'
}

TRAIN_COLOR = EVALUATION_COLORS['train']
VALID_COLOR = EVALUATION_COLORS['valid']
TEST_COLOR = EVALUATION_COLORS['test']
BASELINE_COLOR = EVALUATION_COLORS['baseline']



import importlib
import logging
from pathlib import Path
import requests
import tempfile
from yaml import safe_load

from box import Box
import joblib
import mlflow
import mlflow.sklearn
from mlflow.entities import RunInfo

_logger = logging.getLogger(__name__)


def get_model(run_id: str,
              model_path: str = "model"):
    # Determine how to load the model
    ml_model_url = f"{mlflow.get_tracking_uri()}/get-artifact?path={model_path}%2FMLmodel&run_id={run_id}"
    ml_model = Box(safe_load(requests.get(ml_model_url).content))
    loader_module = ml_model.flavors.python_function.loader_module
    _logger.debug(f"Loader module for the model: {loader_module}")

    # Import the 'load_model' function
    load_model = getattr(importlib.import_module(loader_module), 'load_model')

    # Load the model
    model = load_model(f"runs:/{run_id}/{model_path}")
    return model


def track_model_from_file(local_path: str,
                          artifact_path: str = 'model',
                          experiment_name: str = 'default',
                          run_name: str = None) -> RunInfo:
    _logger.debug("Uploading model to MLflow Server")
    mlflow.set_experiment(experiment_name)
    with mlflow.start_run(run_name=run_name) as run:
        mlflow.log_artifact(local_path=local_path, artifact_path=artifact_path)
        run_info = run.info
    return run_info


def track_model(model,
                artifact_path: str = 'model',
                experiment_name: str = 'default',
                run_name: str = None) -> RunInfo:
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_file = Path(temp_dir) / 'model.joblib'
        _logger.debug(f"Saving model to temporary file: {temp_file}")
        joblib.dump(model, temp_file)
        run_info = track_model_from_file(local_path=temp_file,
                                         artifact_path=artifact_path,
                                         experiment_name=experiment_name,
                                         run_name=run_name)
    return run_info


def track_model_data(run_id: str,
                     params: dict = None,
                     metrics: dict = None,
                     tags: dict = None,
                     artifacts: dict = None,
                     artifact_dirs: dict = None) -> None:
    _logger.debug(f"Tracking model metadata for run_id: {run_id}")
    client = mlflow.tracking.MlflowClient()
    if params is not None:
        for key, value in params.items():
            client.log_param(run_id, key, value)
    if metrics is not None:
        for key, value in metrics.items():
            client.log_metric(run_id, key, value)
    if tags is not None:
        for key, value in tags.items():
            client.set_tag(run_id, key, value)
    if artifacts is not None:
        for local_path, artifact_path in artifacts.items():
            client.log_artifact(run_id, local_path, artifact_path)
    if artifact_dirs is not None:
        for local_dir, artifact_path in artifact_dirs.items():
            client.log_artifacts(run_id, local_dir, artifact_path)


def log_sk_model(sk_model, registered_model_name=None, params=None, tags=None, metrics=None, artifacts=None):
    _logger.info("Logging Scikit-Learn model to MLflow")
    mlflow.sklearn.log_model(sk_model=sk_model,
                             artifact_path='model',
                             conda_env='./environment.yml',
                             registered_model_name=registered_model_name)
    mlflow.log_params(params)
    mlflow.set_tags(tags)
    mlflow.log_metrics(metrics)
    for local_path, artifact_path in artifacts.items():
        _logger.debug(f"Logging artifact to MLflow: {local_path} - {artifact_path}")
        mlflow.log_artifact(local_path, artifact_path)



import logging
from pathlib import Path
from pickle import dump, load

_logger = logging.getLogger(__name__)


def save_pickle(obj, path):
    filepath = Path(path)
    filepath.parent.mkdir(parents=True, exist_ok=True)
    _logger.debug(f"Saving to pickle file: {filepath}")
    with open(filepath, 'wb') as cache_file:
        dump(obj, cache_file)


def load_pickle(path):
    _logger.debug(f"Loading from pickle file: {path}")
    with open(path, 'rb') as cache_file:
        return load(cache_file)



import logging

from sklearn.pipeline import Pipeline

_logger = logging.getLogger(__name__)


def make_pipeline(preprocessor, estimator):
    _logger.debug("Defining the model as a pipeline")
    return Pipeline(steps=[('preprocessor', preprocessor),
                           ('estimator', estimator)])



import logging

import shap

from myautoml.utils.sklearn import get_ct_feature_names
from myautoml.visualisation.evaluation.shap import (
    save_shap_summary, save_shap_dependence_plots, save_shap_summary_bar)

_logger = logging.getLogger(__name__)


def shap_analyse(model, x, temp_dir):
    temp_dir.mkdir(parents=True, exist_ok=True)
    _logger.debug("Performing Shap analysis")
    shap_feature_names = get_ct_feature_names(model.steps[0][1])

    shap_estimator = model.steps[1][1]
    shap_data = model.steps[0][1].transform(x)

    shap_explainer = shap.TreeExplainer(shap_estimator)
    shap_values = shap_explainer.shap_values(shap_data)[1]

    # Computes the baseline value, i.e. expected average shape value for all customers
    tags = {
        'shap_expected_value': shap_explainer.expected_value[1]
    }

    shap_summary_path = save_shap_summary(temp_dir, shap_values, shap_data, shap_feature_names)
    shap_summary_bar_path = save_shap_summary_bar(temp_dir, shap_values, shap_data, shap_feature_names)
    shap_dependence_paths = save_shap_dependence_plots(temp_dir, shap_values, shap_data, shap_feature_names, x,
                                                       model.steps[0][1])

    paths = [shap_summary_path, shap_summary_bar_path, *shap_dependence_paths]
    artifacts = {path: 'shap' for path in paths}

    return tags, artifacts



import logging

_logger = logging.getLogger(__name__)


def flatten_params(params):
    _logger.log(level=logging.NOTSET, msg="Flattening a dictionary of parameters")
    flattened = {}
    for k, v in params.items():
        if isinstance(v, dict):
            # Ignore the higher level and insert the flattened params below this key
            flattened.update(flatten_params(v))
        else:
            flattened[k] = v
    return flattened


def prep_params(params):
    _logger.log(level=logging.NOTSET, msg="Ensuring integer parameters to be integers")
    # Numeric values loaded from a configuration.yaml file tend to have dtype float.
    for p in params.keys():
        if isinstance(params[p], float):
            if params[p] == int(params[p]):
                params[p] = int(params[p])
    return params



import logging
from pathlib import Path

import matplotlib.pyplot as plt

from myautoml.visualisation.colors import EVALUATION_COLORS, TRAIN_COLOR, TEST_COLOR

from . import plot_roc, plot_cum_precision, plot_lift_deciles, plot_precision_recall, plot_prediction_distribution

_logger = logging.getLogger(__name__)


def save_roc_curve(save_dir, data):
    _logger.debug("Plotting the ROC curve")
    save_path = Path(save_dir) / 'roc.png'
    fig, ax = plt.subplots()
    try:
        for label in data.keys():
            plot_roc(ax, data[label]['y'], data[label]['y_pred_proba'],
                     label=label, color=EVALUATION_COLORS[label])
        fig.savefig(save_path)
    except Exception as e:
        _logger.warning(f"Error plotting the ROC curve: {str(e)}")
        save_path = None
    finally:
        plt.close(fig)
    return save_path


def save_precision_recall_curve(save_dir, data):
    _logger.debug("Plotting the Precision-Recall curve")
    save_path = Path(save_dir) / 'precision_recall.png'
    fig, ax = plt.subplots()
    try:
        for label in data.keys():
            plot_precision_recall(ax, data[label]['y'], data[label]['y_pred_proba'],
                                  label=label, color=EVALUATION_COLORS[label],
                                  legend_loc='upper right')
        fig.savefig(save_path)
    except Exception as e:
        _logger.warning(f"Error plotting the Precision-Recall curve: {str(e)}")
        save_path = None
    finally:
        plt.close(fig)
    return save_path


def save_lift_deciles(save_dir, data):
    _logger.debug("Plotting the lift deciles")
    save_path = Path(save_dir) / 'lift_deciles.png'
    fig, ax = plt.subplots()
    try:
        plot_lift_deciles(ax, data['test']['y'], data['test']['y_pred_proba'])
        fig.savefig(save_path)
    except Exception as e:
        _logger.warning(f"Error plotting the lift deciles: {str(e)}")
        save_path = None
    finally:
        plt.close(fig)
    return save_path


def save_cum_precision(save_dir, data):
    _logger.debug("Plotting the cumulative precision curve")
    save_path = Path(save_dir) / 'cum_precision.png'
    fig, ax = plt.subplots()
    try:
        for label in data.keys():
            plot_cum_precision(ax, data[label]['y'], data[label]['y_pred_proba'],
                               label=label, color=EVALUATION_COLORS[label])
        fig.savefig(save_path)
    except Exception as e:
        _logger.warning(f"Error plotting the cumulative precision curve: {str(e)}")
        save_path = None
    finally:
        plt.close(fig)
    return save_path


def save_prediction_distribution(save_dir, data):
    _logger.debug("Plotting the prediction distribution")
    save_path = Path(save_dir) / 'prediction_distribution.png'
    fig, ax = plt.subplots()
    try:
        plot_prediction_distribution(ax, (data['train']['y_pred_proba'], data['test']['y_pred_proba']),
                                     label=('train', 'test'),
                                     color=(TRAIN_COLOR, TEST_COLOR))

        fig.savefig(save_path)
    except Exception as e:
        _logger.warning(f"Error plotting the prediction distribution: {str(e)}")
        save_path = None
    finally:
        plt.close(fig)
    return save_path



import numpy as np

from sklearn.pipeline import Pipeline
from sklearn.feature_extraction.text import _VectorizerMixin
from sklearn.feature_selection._base import SelectorMixin


def get_feature_out(estimator, feature_in):
    if hasattr(estimator, 'get_feature_names'):
        if isinstance(estimator, _VectorizerMixin):
            # handling all vectorizers
            return [f'vec_{f}' \
                    for f in estimator.get_feature_names()]
        else:
            return estimator.get_feature_names(feature_in)
    elif isinstance(estimator, SelectorMixin):
        return np.array(feature_in)[estimator.get_support()]
    else:
        return feature_in


def get_ct_feature_names(ct):
    # handles all estimators, pipelines inside ColumnTransfomer
    # doesn't work when remainder =='passthrough'
    # which requires the input column names.
    output_features = []

    for name, estimator, features in ct.transformers_:
        if name != 'remainder':
            if isinstance(estimator, Pipeline):
                current_features = features
                for step in estimator:
                    current_features = get_feature_out(step, current_features)
                features_out = current_features
            else:
                features_out = get_feature_out(estimator, features)
            output_features.extend(features_out)
        elif estimator == 'passthrough':
            output_features.extend(ct._feature_names_in[features])

    return output_features



import logging

from sklearn.metrics import roc_auc_score, average_precision_score, accuracy_score, f1_score, precision_score, \
    recall_score
from sklearn.model_selection import cross_validate

from myautoml.visualisation.evaluation.binary_classifier import (
    save_roc_curve, save_cum_precision, save_prediction_distribution, save_lift_deciles, save_precision_recall_curve)

_logger = logging.getLogger(__name__)


def evaluate_binary_classifier(model, data, temp_dir, plots='all'):
    _logger.debug(f"Starting evaluation for binary classifier")
    metrics = {}
    for label in data.keys():
        x = data[label]['x']
        y_true = data[label]['y']
        y_pred = data[label]['y_pred'] = model.predict(x)
        y_pred_proba = data[label]['y_pred_proba'] = model.predict_proba(x)[:, 1]

        metrics[f"roc_auc_{label}"] = roc_auc_score(y_true, y_pred_proba)
        metrics[f"average_precision_{label}"] = average_precision_score(y_true, y_pred_proba)
        metrics[f"accuracy_{label}"] = accuracy_score(y_true, y_pred)
        metrics[f"f1_{label}"] = f1_score(y_true, y_pred)
        metrics[f"precision_{label}"] = precision_score(y_true, y_pred)
        metrics[f"recall_{label}"] = recall_score(y_true, y_pred)

    _logger.debug(f"Starting cross-validation for binary classifier")
    scorers = ['roc_auc', 'accuracy', 'f1', 'average_precision', 'precision', 'recall']
    cv_results = cross_validate(estimator=model, X=data['train']['x'], y=data['train']['y'],
                                scoring=scorers, cv=5)
    for scorer in scorers:
        metrics[f"{scorer}_cv"] = cv_results[f"test_{scorer}"].mean()

    artifacts = {}

    if not (plots is None or plots == ""):
        if plots == 'all':
            plots = ['roc', 'pr', 'lift_deciles', 'cum_precision', 'distribution']

        if 'roc' in plots:
            roc_curve_path = save_roc_curve(temp_dir, data)
            artifacts[roc_curve_path] = 'evaluation'

        if 'pr' in plots:
            pr_curve_path = save_precision_recall_curve(temp_dir, data)
            artifacts[pr_curve_path] = 'evaluation'

        if 'lift_deciles' in plots:
            lift_deciles_path = save_lift_deciles(temp_dir, data)
            artifacts[lift_deciles_path] = 'evaluation'

        if 'cum_precision' in plots:
            cum_precision_path = save_cum_precision(temp_dir, data)
            artifacts[cum_precision_path] = 'evaluation'

        if 'distribution' in plots:
            distribution_path = save_prediction_distribution(temp_dir, data)
            artifacts[distribution_path] = 'evaluation'

    return metrics, artifacts



from .__about__ import (
    __title__, __summary__, __uri__, __version__, __author__, __license__, __copyright__,
)

__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__", "__license__", "__copyright__",
]



__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__", "__license__", "__copyright__",
]

__title__ = "myautoml"
__summary__ = "myautoml is a package which provides a framework to automate machine learning"
__uri__ = "https://github.com/myautoml/myautoml"

__version__ = "0.2.4"

__author__ = "Erik Jan de Vries"

__license__ = ""
__copyright__ = "Copyright 2015-2020 {}".format(__author__)



import logging
from pathlib import Path
import getpass


def config_logging(logger=None,
                   log_level=None,
                   log_format=f'%(asctime)s {getpass.getuser()}:%(name)s:%(levelname)s: %(message)s',
                   log_console=True,
                   log_path=None):
    print("Configuring logging")

    # Set logger
    if not isinstance(logger, logging.Logger):
        logger = logging.getLogger(logger)

    if log_level is None:
        log_level = logging.WARNING
    if isinstance(log_level, str):
        log_level = logging.getLevelName(log_level.upper())

    if log_format is None:
        log_format = '%(asctime)s %(name)s:%(levelname)s: %(message)s'

    if log_path is not None:
        log_path = Path(log_path)
        log_path.parent.mkdir(parents=True, exist_ok=True)

    # Generic settings
    logger.setLevel(log_level)
    formatter = logging.Formatter(log_format)

    # Remove existing handlers
    # while logger.hasHandlers():
    while len(logger.handlers) > 0:
        print("Removing handler: {}".format(str(logger.handlers[0])))
        logger.removeHandler(logger.handlers[0])

    # Create console handler
    if log_console:
        ch = logging.StreamHandler()
        ch.setLevel(log_level)
        ch.setFormatter(formatter)
        logger.addHandler(ch)
        logger.info("Added console handler")

    if log_path is not None:
        # Create file handler
        fh = logging.FileHandler(str(log_path), encoding="UTF-8")
        fh.setLevel(log_level)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
        logger.info("Added file handler: {}".format(log_path))


def log_dict(logger, d, name, level=logging.DEBUG):
    """Recursively log the keys and values of a dict."""
    for k, v in d.items():
        if isinstance(v, dict):
            log_dict(logger, v, name=f'{name}.{k}', level=level)
        else:
            logger.log(level, f'{name}.{k}: {v}')



from pathlib import Path
from setuptools import setup, find_packages
import sys

base_dir = Path(__file__).parent
src_dir = base_dir / "src"

# When executing the setup.py, we need to be able to import ourselves, this
# means that we need to add the src/ directory to the sys.path.
sys.path.insert(0, src_dir)

about = {}
with open(src_dir / "myautoml" / "__about__.py") as f:
    exec(f.read(), about)

with open("README.md", "r") as fh:
    long_description = fh.read()

# Optional dependencies
extras = {
    'evaluation': ['shap'],
    'optimisation': ['hyperopt'],
    'tracking': ['mlflow'],
    'visualisation': ['matplotlib', 'seaborn']
}
# Meta dependency groups
extras['all'] = [item for group in extras.values() for item in group]

setup(
    name=about["__title__"],
    version=about["__version__"],

    description=about["__summary__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=about["__uri__"],

    author=about["__author__"],

    project_urls={
        'Documentation': 'https://myautoml.readthedocs.io',
        "Source Code": "https://github.com/myautoml/myautoml",
    },

    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        # "License :: None :: To be determined",
        "Operating System :: OS Independent",
    ],

    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[
        'joblib',
        'numpy',
        'pandas',
        'pyyaml',
        'requests',
        'scikit-learn',

        'python-box',
        'python-dotenv',
        'envyaml',
    ],
    include_package_data=True,

    python_requires='>=3.6',
    extras_require=extras,

    entry_points={
        "console_scripts": [
            "registermodel = myautoml.cli:registermodel"
        ]
    },
)



import logging
import math
from pathlib import Path

import matplotlib.pyplot as plt
from pandas.api.types import is_numeric_dtype, is_bool_dtype
import shap

_logger = logging.getLogger(__name__)


def save_shap_summary(save_dir, shap_values, shap_data, shap_feature_names):
    _logger.debug("Plotting Shap summary diagram")
    save_path = Path(save_dir) / 'shap_summary.png'
    shap.summary_plot(shap_values, shap_data, show=False, feature_names=shap_feature_names)
    fig = plt.gcf()
    fig.set_figwidth(10)
    fig.set_figheight(5)
    fig.savefig(save_path, bbox_inches="tight")
    plt.close(fig)
    return save_path


def save_shap_summary_bar(save_dir, shap_values, shap_data, shap_feature_names):
    _logger.debug("Plotting Shap summary bar diagram")
    save_path = Path(save_dir) / 'shap_summary_bar.png'
    shap.summary_plot(shap_values, shap_data, plot_type='bar', show=False, feature_names=shap_feature_names)
    fig = plt.gcf()
    fig.set_figwidth(10)
    fig.set_figheight(5)
    fig.savefig(save_path, bbox_inches="tight")
    plt.close(fig)
    return save_path


# noinspection DuplicatedCode
def save_shap_dependence_plots(save_dir, shap_values, shap_data, shap_feature_names, display_features, preprocessor=None):
    # Plot dependence plots for each variable in the model
    save_paths = []
    for col in shap_feature_names:
        _logger.debug(f"Plotting dependence plot for {col}")
        save_path = Path(save_dir) / f"shap_dependence_{col}.png"
        shap.dependence_plot(ind=col,
                             shap_values=shap_values,
                             features=shap_data,
                             display_features=display_features[shap_feature_names],
                             # interaction_index=col,
                             interaction_index=None,
                             show=False,
                             feature_names=shap_feature_names)
        fig = plt.gcf()
        fig.set_figwidth(10)
        fig.set_figheight(5)

        # TODO: Check the following hacky solution to fix the x-axis labels in terms of the original data
        if is_numeric_dtype(display_features[col]):
            min_val = display_features[col].min()
            max_val = display_features[col].max()
            if is_bool_dtype(min_val) and min_val:
                min_val = 1
            elif is_bool_dtype(min_val) and not min_val:
                min_val = 0
            else:
                min_val = min_val

            if is_bool_dtype(max_val) and max_val:
                max_val = 1
            elif is_bool_dtype(max_val) and not max_val:
                max_val = 0
            else:
                max_val = max_val

            diff = max_val - min_val
            d = diff
            factors = 0
            while d > 10:
                d = d / 10
                factors = factors + 1

            min_tick = (10 ** factors) * math.ceil(min_val / (10 ** factors))
            max_tick = (10 ** factors) * math.floor(max_val / (10 ** factors))

            ticks_values = [x for x in range(min_tick, max_tick + 1, 10 ** factors)]

            # TODO: This specific solution for age should be dealt with differently
            if col == 'age':
                ticks_values = [x for x in range(min_tick, max_tick + 1, 10)]
            # df_ticks_values = pd.DataFrame(ticks_values, columns=[col])

            df_ticks_values = display_features.head(len(ticks_values)).copy()
            df_ticks_values[col] = ticks_values

            ticks = preprocessor.transform(df_ticks_values)
            col_idx = shap_feature_names.index(col)
            ticks = ticks[:, col_idx]

            plt.xticks(ticks, ticks_values)

        plt.axhline(y=0, color='black', linestyle='--', alpha=0.4)
        fig.savefig(save_path, bbox_inches="tight")
        plt.close(fig)
        save_paths.append(save_path)
    return save_paths



import click
import mlflow

from myautoml.utils.mlflow import track_model_from_file, track_model_data, register_model


@click.command()
@click.option('--experiment_name', '-e', help='Experiment name in MLflow')
@click.option('--model_path', '-m', help='Path to the trained model file')
@click.option('--model_name', '-n', help='Name to register the model with')
def registermodel(experiment_name, model_path, model_name):
    if experiment_name is None:
        experiment_name = "Default"
    if model_path is not None:
        click.secho("Uploading model to MLflow Server", fg="green", bold=True)
        run_info = track_model_from_file(local_path=model_path,
                                         experiment_name=experiment_name)
        track_model_data(run_info.run_id, tags={"cli_upload": True})
        experiment_run_url = f"{mlflow.get_tracking_uri()}/#/experiments/{run_info.experiment_id}/runs/{run_info.run_id}"

        if model_name is not None:
            click.secho("Registering model with MLflow Server", fg="green", bold=True)
            model_version = register_model(run_id=run_info.run_id,
                                           model_name=model_name)
            registered_model_url = f"{mlflow.get_tracking_uri()}/#/models/{model_name}/versions/{model_version.version}"
        else:
            click.secho(f"No name specified to register the model!", fg="white", bold=True)
            registered_model_url = ""

        click.echo(f"Experiment run URL: {experiment_run_url}")
        click.echo(f"Registered model URL: {registered_model_url}")

    else:
        click.secho(f"No model specified to upload!", fg="red", bold=True)
